#ifndef TENSOR_H
#define TENSOR_H
#include "matrix.h"
#include <iostream>
#include <fstream>
#include <string>
#include <map>
#define Pi M_PI
#include <fenv.h> 
#include <assert.h>
#include <utility>
#include "boost/shared_ptr.hpp"
#include <math.h>

using namespace std::rel_ops;

inline Real sqr(Real x) { return x*x; }

void ComplexEigenvalues(const Matrix& re, const Matrix& im, Vector& evals,
	                                Matrix& revecs, Matrix& ievecs);

typedef int int9[9];

extern bool printdat;
extern Vector lastd;
extern ofstream big_op_file;
extern bool writeops;


template<class T, class Op> void for_all(T& a, Op f)
        { for_each(a.begin(),a.end(),f); }

inline void set8(int9 *n, int i1, int i2, int i3, int i4, int i5, int i6, int i7, int i8)
    {
#ifdef SUPERCHECK
    assert(i1 > 0); assert(i2 > 0); assert(i3 > 0); assert(i4 > 0); 
    assert(i5 > 0); assert(i6 > 0); assert(i7 > 0); assert(i8 > 0);
#endif
    (*n)[1] = i1; (*n)[2] = i2; (*n)[3] = i3; (*n)[4] = i4;
    (*n)[5] = i5; (*n)[6] = i6; (*n)[7] = i7; (*n)[8] = i8;
    }

class Tensor
    {
public:
    int d,n[9];
    Vector dat;

    int ind(int i1 = 1,int i2 = 1,int i3 = 1,int i4 = 1,int i5 = 1,
	    int i6 = 1,int i7 = 1, int i8 = 1) const
	{
#ifdef SUPERCHECK
	assert(i1 > 0); assert(i2 > 0); assert(i3 > 0); assert(i4 > 0);
	assert(i5 > 0); assert(i6 > 0); assert(i7 > 0); assert(i8 > 0);
	assert(i1 <= n[1]); assert(i2 <= n[2]); assert(i3 <= n[3]); assert(i4 <= n[4]);
	assert(i5 <= n[5]); assert(i6 <= n[6]); assert(i7 <= n[7]); assert(i8 <= n[8]);
#endif
	return (((((((i8-1)*n[7]+i7-1)*n[6]+i6-1)*n[5]+i5-1)*n[4]+i4-1)
		    *n[3]+i3-1)*n[2]+i2-1)*n[1]+i1;
	}
    Real& operator()(int i1 = 1,int i2 = 1,int i3 = 1,int i4 = 1,int i5 = 1,
	    int i6 = 1,int i7 = 1, int i8 = 1) 
	{
	return dat(ind(i1,i2,i3,i4,i5,i6,i7,i8));
	}
    Real operator()(int i1 = 1,int i2 = 1,int i3 = 1,int i4 = 1,int i5 = 1,
	    int i6 = 1,int i7 = 1, int i8 = 1) const
	{
	return dat(ind(i1,i2,i3,i4,i5,i6,i7,i8));
	}

    Tensor(int dd = 0, int i1 = 1,int i2 = 1,int i3 = 1,int i4 = 1,int i5 = 1,
	    int i6 = 1,int i7 = 1, int i8 = 1) : d(dd), dat(i1*i2*i3*i4*i5*i6*i7*i8)
	{
	set8(&n,i1,i2,i3,i4,i5,i6,i7,i8);
	if(d < 0)	// it will calculate
	    for(int i = 8; i >= 0; i--)
		if(n[i] > 1) { d = i; break; }
	if(dat.Length() > 0) dat = 0.0;
	}

    Tensor(int nn1,int nn2,Real a) : d(2), dat(nn1*nn2)
	{
	dat = 0.0;
	set8(&n,nn1,nn2,1,1,1,1,1,1);
	int nn = min(nn1,nn2);
	for(int i = 1; i <= nn; i++)
	    (*this)(i,i) = a;
	}
	
    void Resize(int dd, int nn1 = 1,int nn2 = 1,int nn3 = 1,int nn4 = 1,int nn5 = 1,
	    int nn6 = 1,int nn7 = 1, int nn8 = 1) 
	{ 
	dat.ReDimension(nn1*nn2*nn3*nn4*nn5*nn6*nn7*nn8);  
	set8(&n,nn1,nn2,nn3,nn4,nn5,nn6,nn7,nn8);
	d = dd;
	if(d < 0)	// it will calculate
	    for(int i = 8; i >= 0; i--)
		if(n[i] > 1) { d = i; break; }
	}
    Tensor& operator+=(const Tensor& o) { dat += o.dat; return *this; }
    Tensor& operator*=(Real fac) { dat *= fac; return *this; }
    Tensor operator+(const Tensor& o) { Tensor res(*this); res += o; return res; }
    Tensor operator*(Real fac) { Tensor res(*this); res *= fac; return res; }
    Tensor(const MatrixRef& mat)
	{
	Resize(2,mat.Nrows(),mat.Ncols());
	for(int i = 1; i <= n[1]; i++)
	    for(int j = 1; j <= n[2]; j++)
		 operator()(i,j) = mat(i,j);
	}

    friend ostream & operator << (ostream & s, const Tensor & t);
    };

inline Tensor operator*(Real fac,const Tensor& t) { Tensor res(t); res *= fac; return res; }
class Counter
    {
public:
    int n[9],i[9],ind;
    void init(int a)
	{
	i[0]=i[1]=i[2]=i[3]=i[4]=i[5]=i[6]=i[7]=i[8]=a;
	ind = 1;
	}
    Counter(const Tensor& t)
{
	for(int j = 1; j <= 8; j++)
	    n[j] = t.n[j];
	init(1);
	}
    Counter()
	{
	init(0);
	n[0]=n[1]=n[2]=n[3]=n[4]=n[5]=n[6]=n[7]=n[8]=1;
	}
    Counter& operator++()
	{
	ind++;
	i[1]++;
	if(i[1] > n[1])
	    for(int j = 2; j <= 8; j++)
		{
		i[j-1] = 1;
		i[j]++;
		if(i[j] <= n[j]) break;
		}
	if(i[8] > n[8])
	    init(0);
	return *this;
	}
    bool operator==(const Counter& other) const
	{
	for(int j = 1; j <= 8; j++)
	    if(i[j] != other.i[j])
		return false;
	return true;
	}
    bool operator!=(const Counter& other) const
	{
	return !(*this == other);
	}
    Real& val(Tensor& x)
	{
	return x(i[1],i[2],i[3],i[4],i[5],i[6],i[7],i[8]);
	}
    Real val(const Tensor& x)
	{
	return x(i[1],i[2],i[3],i[4],i[5],i[6],i[7],i[8]);
	}
    static Counter *pend;
    static Counter& done;
    };
inline ostream& operator<<(ostream& s, const Counter& c)
    {
    bool first = true;
    s << "(";
    for(int i = 1; i <= 8; i++)
	if(c.n[i] > 1)
	    {
	    if(!first)
		s << ",";
	    s << c.i[i];
	    first = false;
	    }
    s << ")";
    return s;
    }
#ifdef THIS_IS_MAIN
Counter * Counter::pend = new Counter;
Counter & Counter::done(*pend);
#endif                                  
class Permutation // Tell where each index will go, p(2,1,3) says 1 -> 2, 2 -> 1, 3 -> 3
    {
public:
    int ind[9];
    Permutation(int i1 = 1, int i2 = 2, int i3 = 3, int i4 = 4, int i5 = 5, int i6 = 6,
	    int i7 = 7,  int i8 = 8)
	{
	set8(&ind,i1,i2,i3,i4,i5,i6,i7,i8);
	}
    void check(int d)
	{
	for(int i = 1; i <= d; i++)
	    if(ind[i] > d || ind[i] < 1)
		error("bad Permutation level 1");
	for(int i = 1; i <= d; i++)
	    for(int j = 1; j <= d; j++)
		if(ind[i] == ind[j] && i != j)
		    error("bad Permutation level 2");
	}
    };

inline ostream& operator<<(ostream& s, const Permutation& p)
    {
    for(int i = 1; i <= 8; i++)
	s << p.ind[i];
    return s;
    }

void Reshape(const Tensor& x, Permutation p, Tensor& res);

void Product(const Tensor& x, const Tensor& y, Tensor& res);

inline Tensor operator*(const Tensor& x,const Tensor& y)
    {
    Tensor& xx(const_cast<Tensor&>(x));
    Tensor& yy(const_cast<Tensor&>(y));
    Tensor res;
    Product(xx,yy,res);
    return res;
    }

enum IndexType { Link,  Site, Factored, Link2,ReIm};
static const char * indextypename[] = {"Link","Site", "Factored", "Link2","ReIm"};
enum PrimeType { primeLink, primeSite, primeBoth, primeNone};
class prime
    {
public:
    PrimeType t;
    prime(PrimeType tt) : t(tt) {}
    bool operator==(const prime& other) { return t == other.t; }
    };
static prime PrimeLink(primeLink), PrimeSite(primeSite), 
    PrimeBoth(primeBoth), PrimeNone(primeNone);

inline string putprimes(string s, int plev = 0)
    {
    for(int i = 1; i <= plev; i++)
	s += "\'";
    return s;
    }
inline string nameindex(IndexType it, int plev = 0)
    {
    return putprimes(string(indextypename[(int)it]),plev);
    }

enum Imaker {makeReIm,makeReImP,makeReImPP};

extern boost::shared_ptr<string> nullname;

class Index
    {
    string sname;
    boost::shared_ptr<string> psname;
    IndexType _type;
public:
    int ind, primelevel, m_;
    Real ur;
    IndexType type() const { return _type; }

    static int indcount;

    Real unique_Real() const { return ur; }
    void set_unique_Real()
	{
	ur = sin(ind * sqrt(1.0/7.0) + primelevel * sqrt(1.0 / 11.0)
		+ ((int)_type) * sqrt(1.0 / 13.0));
	}
    string name() const  { return putprimes(*psname,primelevel); }
    void setname(string s) 
	{ 
        psname.reset(new string(s));
        //sname = s; 
	} 
    Index(IndexType it, string ss = " ", int mm = 2) 
		: psname(new string(ss)), _type(it), ind(indcount++), primelevel(0), m_(mm)
	{ 
	if(it == ReIm) error("bad call to create Index with ReIm");
	set_unique_Real();
	}

    Index() : psname(nullname),  _type(Site), ind(0), primelevel(0), m_(0), ur(0.0) {}
    Index(Imaker im) : psname(nullname), _type(ReIm), ind(1), m_(2)
	{
	if(im == makeReIm)
	    setname("ReIm"), primelevel = 0;
	if(im == makeReImP)
	    setname("ReImP"),  primelevel = 1;
	if(im == makeReImPP)
	    setname("ReImPP"),  primelevel = 2;
	set_unique_Real();
	}

    int m() const { return m_; }
    void setm(int mm) { m_ = mm; }

    // rel_ops defines the other comparisons based on == and <
    bool operator==(const Index& other) const 
	{ 
	return ur == other.ur;
	//return type==other.type && ind == other.ind && primelevel == other.primelevel; 
	}
    bool operator<(const Index& other) const 
	{ 
	return ur < other.ur;
	//int ty1 = static_cast<int>(type);
	//int ty2 = static_cast<int>(other.type);
	//return ty1 < ty2 || (ty1 == ty2 && (ind < other.ind || 
	//	    (ind == other.ind && primelevel < other.primelevel))); 
	}
    bool noprime_equals(const Index& other) const
	{
	return type()==other.type() && ind == other.ind;
	}
    void mapprime(int plevold, int plevnew, prime p = PrimeBoth)
	{
	if(type() == ReIm) return;
	if(primelevel != plevold) return;
	else if( p.t == primeBoth 
	    || (type() == Site && p.t == primeSite) 
	    || (type() == Link && p.t == primeLink) 
	    || (type() == Link2 && p.t == primeLink) )
	    {
	    primelevel = plevnew;
	    set_unique_Real();
	    }
	}
    void doprime(prime p)
	{
	if(type() == ReIm) return;
	if( p.t == primeBoth 
	    || (type() == Site && p.t == primeSite) 
	    || (type() == Link && p.t == primeLink) 
	    || (type() == Link2 && p.t == primeLink) )
	    {
	    primelevel++;
	    set_unique_Real();
	    }
	}
    Index primed() const
	{
	return Index(primeBoth,*this);
	}
    void noprime() { primelevel = 0; set_unique_Real(); }
    Index(prime p,const Index& other, int primeinc = 1) 
	: _type(other._type), ind(other.ind), primelevel(other.primelevel),
				m_(other.m_)
	{
	setname(other.name());
	for(int i = 1; i <= primeinc; i++)
	    doprime(p);
	set_unique_Real();
	}
    };

inline string nameint(string f,int ix)
    {
    stringstream ss;
    ss << f << ix;
    return ss.str();
    }

extern Index IndReIm, IndReImP, IndReImPP;

inline ostream & operator << (ostream & s, const Index & t)
    {
    if(t.name() != "" && t.name() != " ")
	s << t.name() << "/";
    return s << nameindex(t.type(),t.primelevel) << "-" << t.ind << ":" << t.m();
    }

class IV
    {
public:
    Index ind;
    int i;
    IV(Index index, int ii) : ind(index), i(ii) {}
    IV() {}
    };

enum ITmaker {makeComplex_1,makeComplex_i};

class ITensor : public Tensor	// Index Tensor
    {
    Index _index[9];	// max dim is 8
public:
    Real logfac, ur;
    static int maxdim;
    Real unique_Real() const { return ur; }	// depends on indices only, unordered
    //Index& index(int i) { return _index[i]; }
    void add_index(const Index& j) { d += 1; _index[d] = j; resize_from_m(); set_unique_Real();}
    void set_index(int i, const Index& j) { _index[i] = j; set_unique_Real();}
    const Index& index(int i) const { return _index[i]; }
    void set_unique_Real()
	{
	ur = 0.0;
	for(int j = 1; j <= d; j++)
	    ur += _index[j].unique_Real();
	}
    ITensor(Index i1,Index i2,Real a) 
	    : Tensor(i1.m(),i2.m(),a), logfac(0.0)
	{ _index[1] = i1; _index[2] = i2; set_unique_Real(); }
    ITensor() : logfac(0.0), ur(0.0) {}
    ITensor(Index i1, Real val = 0.0) : Tensor(1,i1.m()), logfac(0.0)
	{ _index[1] = i1; dat = val; set_unique_Real(); }
    ITensor(Index i1,Index i2) : Tensor(2,i1.m(),i2.m()), logfac(0.0)
	{ _index[1] = i1; _index[2] = i2;  set_unique_Real(); }
    ITensor(Index i1,Index i2,Index i3) : Tensor(3,i1.m(),i2.m(),i3.m()), logfac(0.0)
	{ _index[1] = i1; _index[2] = i2; _index[3] = i3;  set_unique_Real(); }
    ITensor(Index i1,Index i2,Index i3,Index i4) 
		: Tensor(4,i1.m(),i2.m(),i3.m(),i4.m()), logfac(0.0)
	{ _index[1] = i1; _index[2] = i2; _index[3] = i3; _index[4] = i4;  set_unique_Real(); }
    ITensor(Index i1,Index i2,Index i3,Index i4,Index i5) 
		: Tensor(5,i1.m(),i2.m(),i3.m(),i4.m(),i5.m()), logfac(0.0)
	{ _index[1] = i1; _index[2] = i2; _index[3] = i3; _index[4] = i4; _index[5] = i5;  set_unique_Real(); }
	
	
    // separately fill in indices, not the best way
//    ITensor(int nn1,int nn2,int nn3,int nn4,int nn5 = 1,
	    //int nn6 = 1,int nn7 = 1, int nn8 = 1) 
	//: Tensor(-1,nn1,nn2,nn3,nn4,nn5,nn6,nn7,nn8), logfac(0.0) {} 

    ITensor(ITmaker itm) : Tensor(1,2), logfac(0.0)
	{
	_index[1] = IndReIm;
	if(itm == makeComplex_1) dat(1) = 1.0;
	if(itm == makeComplex_i) dat(2) = 1.0;
	set_unique_Real();
	}
    void product(const ITensor& other,ITensor& res) const;
    ITensor operator*(const ITensor& other) const;
    ITensor operator/(const ITensor& other) const;  // actually noncontracting product
    void doprime(prime p)
	{
	for(int j = 1; j <= d; j++)
	    _index[j].doprime(p);
	set_unique_Real();
	}
    ITensor(prime p,const ITensor& other)
	: Tensor(other), _index(other._index), logfac(other.logfac)
	{
	doprime(p);
	set_unique_Real();
	}
    void mapprime(int plevold, int plevnew, prime p = PrimeBoth)
	{
	for(int j = 1; j <= d; j++)
	    _index[j].mapprime(plevold,plevnew,p);
	set_unique_Real();
	}
    ITensor primelink() const { return ITensor(primeLink,*this); }
    ITensor primesite() const { return ITensor(primeSite,*this); }
    ITensor primeind(Index k) const 
	{ 
	ITensor res(*this);
	res.mapindex(k,k.primed());
	return res;
	}
    ITensor primeind(Index k, Index k2) const 
	{ 
	ITensor res(*this);
	res.mapindex(k,k.primed());
	res.mapindex(k2,k2.primed());
	return res;
	}
    ITensor primeboth() const { return ITensor(primeBoth,*this); }
    ITensor primed(prime p = PrimeBoth) const { return ITensor(p,*this); }

    void Assign(const ITensor& other); // Assume *this and other have same indices but different order.
    		// Copy other into *this, without changing the order of indices in either
    		// operator= would put the order of other into *this

    ITensor operator+(const ITensor& o) const { ITensor res(*this); res += o; return res; }
    ITensor& operator+=(const ITensor& o);
    ITensor& operator*=(Real fac) { dat *= fac; return *this; }
    ITensor operator*(Real fac) { ITensor res(*this); res *= fac; return res; }
    void mapindex(Index i1, Index i2)
	{
	for(int j = 1; j <= d; j++)
	    if(_index[j] == i1) _index[j] = i2;
	set_unique_Real();
	}
    void noprime()
	{
	for(int j = 1; j <= d; j++)
	    _index[j].noprime();
	set_unique_Real();
	}
    void mapindex(Index i1, Index i2, Index i3, Index i4)
	{
	mapindex(i1,i2);
	mapindex(i3,i4);
	}
    void mapindex(Index i1, Index i2, Index i3, Index i4, Index i5, Index i6)
	{
	mapindex(i1,i2);
	mapindex(i3,i4);
	mapindex(i5,i6);
	}
    /*
    void maptype(IndexType t, IndexType t2)
	{
	for(int j = 1; j <= d; j++)
	    if(index[j].type() == t) index[j].type() = t2;
	}
	*/
    Index findtype(IndexType t) const
	{
	for(int j = 1; j <= d; j++)
	    if(_index[j].type() == t) return _index[j];
	error("couldnt find type 832489");
	}
    bool findtype(IndexType t,Index& i) const
	{
	for(int j = 1; j <= d; j++)
	    if(_index[j].type() == t)
		{
		i = _index[j];
		return true;
		}
	return false;
	}
    int findindex(Index i) const
	{
	for(int j = 1; j <= d; j++)
	    if(_index[j] == i) return j;
	return 0;
	}

    bool notin(Index i) const { return findindex(i)==0; }

    int nindex(Index i) const
	{
	for(int j = 1; j <= d; j++)
	    if(_index[j] == i) return n[j];
	return 0;
	}
    void resize_from_m()
	{
	int len = 1;
	for(int j = 1; j <= d; j++)
	    {
	    n[j] = _index[j].m();
	    len *= _index[j].m();
	    if(n[j] == 0)
		{
		cout << *this << endl;
		error("bad size in resize from m");
		}
	    }
	dat.ReDimension(len);
	dat = 0.0;
	}
    void resize_from_m(Index ia)
	{
	for(int j = 1; j <= d; j++)
	    if(_index[j] == ia)
		_index[j].setm(ia.m());
	resize_from_m();
	}
    void resize_from_m(Index ia,Index ib)
	{
	for(int j = 1; j <= d; j++)
	    if(_index[j] == ia)
		_index[j].setm(ia.m());
	for(int j = 1; j <= d; j++)
	    if(_index[j] == ib)
		_index[j].setm(ib.m());
	resize_from_m();
	}
    Real norm() const { return Norm(dat) * exp(logfac); }
    Real lognorm() const { return log(Norm(dat) + 1.0e-100) + logfac; }
    void donormlog()
	{
	Real f = Norm(dat);
	if(f != 0.0)
	    dat *= 1.0/f, logfac += log(f);
	}
    void normlogto(Real newlogfac)
	{
	Real dellogfac = newlogfac - logfac;
	logfac = newlogfac;
	if(dellogfac > 30.0)
	    dat = 0.0;
	else
	    dat *= exp(-dellogfac);
	}

    Real& val5(int i1,int i2,int i3,int i4,int i5)
	{ return dat(((((i5-1)*n[4]+i4-1)*n[3]+i3-1)*n[2]+i2-1)*n[1]+i1); }
    Real& val4(int i1,int i2,int i3,int i4)
	{ return dat((((i4-1)*n[3]+i3-1)*n[2]+i2-1)*n[1]+i1); }
    Real& val3(int i1,int i2,int i3)
	{ return dat(((i3-1)*n[2]+i2-1)*n[1]+i1); }
    Real& val2(int i1,int i2)
	{ return dat((i2-1)*n[1]+i1); }

    Real& val(IV iv1)
	{
	if(d != 1) error("wrong d in val" );
	if(_index[1] != iv1.ind) error("wrong ind 1 in val" );
	return dat(iv1.i);
	}
    Real& val(IV iv1, IV iv2)
	{
	if(d != 2) error("wrong d in val" );
	int ia[9], ja[9];
	Index ind[9];
	ia[1] = iv1.i; ia[2] = iv2.i;
	ind[1] = iv1.ind; ind[2] = iv2.ind;
	for(int k = 1; k <= d; k++)	// loop over indices of this ITensor
	    {
	    bool gotit = false;
	    for(int m = 1; m <= d; m++)  // loop over the given indices
		if(_index[k] == ind[m])
		    gotit = true, ja[k] = ia[m];
	    if(!gotit)
		error("couldnt get index in val");
	    }
	return val2(ja[1],ja[2]);
	}
    Real& val(IV iv1, IV iv2, IV iv3)
	{
	if(d != 3) error("wrong d in val" );
	int ia[9], ja[9];
	Index ind[9];
	ia[1] = iv1.i; ia[2] = iv2.i; ia[3] = iv3.i;
	ind[1] = iv1.ind; ind[2] = iv2.ind; ind[3] = iv3.ind;
	for(int k = 1; k <= d; k++)	// loop over indices of this ITensor
	    {
	    bool gotit = false;
	    for(int m = 1; m <= d; m++)  // loop over the given indices
		if(_index[k] == ind[m])
		    gotit = true, ja[k] = ia[m];
	    if(!gotit)
		error("couldnt get index in val");
	    }
	return val3(ja[1],ja[2],ja[3]);
	}
    Real& val(IV iv1, IV iv2, IV iv3, IV iv4)
	{
	if(d != 4) error("wrong d in val" );
	int ia[9], ja[9];
	Index ind[9];
	ia[1] = iv1.i; ia[2] = iv2.i; ia[3] = iv3.i; ia[4] = iv4.i;
	ind[1] = iv1.ind; ind[2] = iv2.ind; ind[3] = iv3.ind; ind[4] = iv4.ind;
	for(int k = 1; k <= d; k++)	// loop over indices of this ITensor
	    {
	    bool gotit = false;
	    for(int m = 1; m <= d; m++)  // loop over the given indices
		if(_index[k] == ind[m])
		    gotit = true, ja[k] = ia[m];
	    if(!gotit)
		error("couldnt get index in val");
	    }
	if(0)cout << iv1.ind SP iv1.i << "   " << iv2.ind SP iv2.i
		    << "   " << iv3.ind SP iv3.i<< "   " << iv4.ind SP iv4.i << endl;
	if(0)cout << ja[1] SP ja[2] SP ja[3] SP ja[4] << endl;
	return val4(ja[1],ja[2],ja[3],ja[4]);
	}
    friend ostream & operator << (ostream & s, const ITensor & t);

    void getMatrix11(Index i1, Index i2, Matrix& res) const;
    void fromMatrix11(Index i1, Index i2, const Matrix& res);
// group i1,i2; i3,i4
    void getMatrix22(Index i1, Index i2, Index i3, Index i4,Matrix& res) const;
    void fromMatrix22(Index i1, Index i2, Index i3, Index i4,const Matrix& res);
    void getMatrix21(Index i1, Index i2, Index i3, Matrix& res) const;
    void getMatrix12(Index i1, Index i2, Index i3, Matrix& res) const;
    void fromMatrix21(Index i1, Index i2, Index i3, const Matrix& res);
    void fromMatrix12(Index i1, Index i2, Index i3, const Matrix& res);

    void PrimesToBack();
    void SplitReIm(ITensor& re, ITensor& im) const;
    void AssignToVec(VectorRef v) const
	{
	if(dat.Length() != v.Length())
	    error("ITensor::AssignToVec bad size");
	//cout << "dat, exp(logfac) are " << dat SP exp(logfac) << endl;
	v = dat* exp(logfac);
	//cout << "v is " << v << endl;
	}
    void AssignFromVec(VectorRef v)
	{
	if(dat.Length() != v.Length())
	    error("ITensor::AssignToVec bad size");
	logfac = 0.0;
	dat = v;
	}
    };
void Reshape(const ITensor& x, Permutation p, ITensor& res);
void IndexToFront(ITensor& A, Index i);
void IndexToFront(ITensor& A, Index i, Index j);
inline ITensor operator*(Real fac,const ITensor& t) { ITensor res(t); res *= fac; return res; }
inline Real Dot(const ITensor& x, const ITensor& y)
    {
    if(x.d != y.d) 
	{
	cout << x << endl;
	cout << y << endl;
	error("bad Dot");
	}
    ITensor res(x*y);
    if(res.d != 0) 
	{
	cout << x << endl;
	cout << y << endl;
	error("bad Dot 2");
	}
    return res(1) * exp(res.logfac);
    }

extern ITensor Complex_1, Complex_i;

inline ITensor conj(const ITensor& A)
    {
    if(A.findindex(IndReIm) == 0)
	return A;
    ITensor r,i;
    A.SplitReIm(r,i);
    i *= -1.0;
    return r * Complex_1 + Complex_i * i;
    }
inline ITensor deprimed(const ITensor& A)
    {
    ITensor a = A;
    a.noprime();
    return a;
    }

inline Real RDot(const ITensor& x, const ITensor& y)
    {
    if(x.findindex(IndReIm) > 0)
	{
	ITensor res = conj(x) * y;
	if(res.d != 1)
	    error("Bad RDot 234234");
	return res(1) * exp(res.logfac);
	}
    else if(y.findindex(IndReIm) > 0)
	{
	ITensor res = x * y;
	if(res.d != 1)
	    error("Bad RDot 37298789");
	return res(1) * exp(res.logfac);
	}
    if(x.d != y.d) 
	{
	cout << x << endl;
	cout << y << endl;
	error("bad Dot");
	}
    ITensor res(x*y);
    if(res.d != 0) 
	{
	cout << x << endl;
	cout << y << endl;
	error("bad Dot 2");
	}
    return res(1) * exp(res.logfac);
    }

Index index_in_common(const ITensor& A, const ITensor& B, IndexType t);

/*
inline Real CDot(const ITensor& x, const ITensor& y)
    {
    if(x.d != y.d) 
	{
	cout << x << endl;
	cout << y << endl;
	error("bad CDot");
	}
    ITensor xi,xr,yi,yr;
    x.SplitReIm(xr,xi);
    y.SplitReIm(yr,yi);
    return Dot(xr,yr) + Dot(xi,yi);
    }
*/

// group i1,i2; i3,i4
ITensor Inverse(const ITensor& Q,Index i1, Index i2, Index i3, Index i4);

/*
Combine several indices into one, use * to convert tensors efficiently
   \
    \
  ---C====
    /
   /

*/
class Combiner
    {
public:
    Index left[9];	// max dim is 8
    Index _right;
    static Index spec;
    int dl;
    Combiner() : dl(0) {}
    Combiner(Index& r, 
	    Index l1 = spec, Index l2 = spec, Index l3 = spec, Index l4 = spec, 
	    Index l5 = spec, Index l6 = spec, Index l7 = spec, Index l8 = spec )
	{
        left[0] = spec;
        left[1] = l1; left[2] = l2; left[3] = l3; left[4] = l4;
        left[5] = l5; left[6] = l6; left[7] = l7; left[8] = l8;
        for(dl = 8; dl >= 1; dl--)
            if(left[dl] != spec)
            break;
        int dim = 1;
        for(int i = 1; i <= dl; i++)
            dim *= left[i].m();
        r.setm(dim);
        _right = r;
	}
    Index right() const { return _right; }
    void addleft(Index& r, Index l) 	// Include another left index
	{ 
        left[++dl] = l; 
        int m = 1;
        for(int i = 1; i <= dl; i++)
            m *= left[i].m();
        r.setm(m);
        _right = r;
	}
    int findindex(Index i) const
	{
	for(int j = 1; j <= dl; j++)
	    if(left[j] == i) return j;
	return 0;
	}
    };
ITensor operator*(const ITensor& t, const Combiner& c);
inline ostream & operator << (ostream & s, const Combiner & c)
    {
    s << c.right() << endl;
    for(int i = 1; i <= c.dl; i++)
	s << "	" << c.left[i] << endl;
    return s;
    }

inline ITensor operator*(const Combiner& c, const ITensor& t)
    {
    return t * c;
    }

inline Real getentropy()
    {
    int n = lastd.Length();
    Real nor = 0.0;
    for(int i = 1; i <= n; i++)
	nor += lastd(i)*lastd(i);
    Vector dd = lastd * (1.0 / sqrt(nor));
    Real S = 0.0;
    for(int i = 1; i <= n; i++)
	if(dd(i) > 0.0)
	    S += -dd(i)*dd(i) * log(dd(i)*dd(i));
    return S;
    }

enum SweepScheme {ramp_m, fixed_m, fixed_cutoff};

#include <vector>

inline void sweepnext(int &l, int &ha, int N)
    {
    if(ha == 1)
	{
	if(++l == N) 
	    l = N-1, ha = 2;
	return;
	}
    if(--l == 0) ha = 3;
    }

class Sweeps
    {
public:
    vector<int>  Maxm, Niter;
    vector<Real> Cutoff;
    int Nsweep;
    Sweeps(SweepScheme sch, int nsw, int _minm, int _maxm, Real _cut)
	    : Maxm(nsw+1), Niter(nsw+1,4), Cutoff(nsw+1), Nsweep(nsw)
	{
	if(sch == ramp_m)
	    {
	    for(int s = 1; s <= Nsweep; s++)
		Cutoff[s] = _cut, Maxm[s] = (int)(_minm + (s-1.0)/nsw * (_maxm - _minm));
	    }
	else if(sch == fixed_m || sch == fixed_cutoff)
	    for(int s = 1; s <= Nsweep; s++)
		Cutoff[s] = _cut, Maxm[s] = _maxm;
	for(int s = 1; s <= min(Nsweep,4); s++)
	    Niter[s] = 10 - s;
	}
    Real cutoff(int sw) const { return Cutoff[sw]; }
    int maxm(int sw) const { return Maxm[sw]; }
    int nsweep() const { return Nsweep; }
    int niter(int sw) const { return Niter[sw]; }
    };


#ifdef THIS_IS_MAIN
boost::shared_ptr<string> nullname(new string(" "));
int Index::indcount = 0;
int ITensor::maxdim = 8;
bool printdat = true;
bool writeops = false;
ofstream big_op_file;
Index Combiner::spec = Index(Link,"spec",2);
Index IndReIm(makeReIm);
Index IndReImP(makeReImP);
Index IndReImPP(makeReImPP);
ITensor Complex_1(makeComplex_1), Complex_i(makeComplex_i);
Vector lastd(1);
#endif
#endif
